{: Unit con procedures generales que comparten Proyectos.
@author     Germn Estvez
@cat        Procedimientos generales a la aplicacin
}
{
=========================================================================

  UCommonProcedures.pas

  Tipo: Unit

========================================================================
  Historia de las Versiones
------------------------------------------------------------------------

  07/08/2014   * Creacin.

  Unit con procedimientos sin dependencias para compartir entre proyectos

=========================================================================

  Errores detectados no corregidos

=========================================================================
}
unit UCommonProcedures;




//=========================================================================
//
// I N T E R F A C E
//
//=========================================================================
interface

{$I jedi.inc}

uses
  Classes, Windows, ShellAPI,
{$IFDEF DELPHIXE_UP}
  FMX.Controls;
{$ELSE}
  controls;
{$ENDIF}


type
{$IFNDEF WIN32}
   TStrLen = Byte;
{$ELSE}
  TStrLen = Integer;
{$ENDIF}

  {: Tipo para un procedimiento de log.}
  TLogProcedure = procedure(AMsg: string; AWithTime:boolean=True) of Object;

// Ficheros INI de configuracin   =============================================
//: Guarda un array de valores en el fichero de configuracion.
procedure SaveOptionValues(AIniFilePath, ASectionName:string;
                           AName, AValue:array of String); overload;
procedure SaveOptionValues(AIniFilePath, ASectionName:string;
                           AName:string; AValue:TStrings); overload;
{: Recupera un valor almacenado en el fichero de configuracin.}
function RestoreOptionValue(AIniFilePath, ASectionName:string;
                            AName, ADefault:string):String; overload;
function RestoreOptionValue(AIniFilePath, ASectionName:string;
                            AName:string; ADefault:Integer): Integer; overload;
function RestoreOptionValue(AIniFilePath, ASectionName:string;
                            AName:string; ADefault:boolean): boolean; overload;
procedure RestoreOptionValue(AIniFilePath, ASectionName:string;
                            AName:string; var outStrings:TStrings); overload;
//: Devuelve si est guardada determinada opcin
function OptionExist(AIniFilePath, ASectionName:string; AName:string):Boolean;


// FECHAS ======================================================================
//: devuelve el primer dia del mes y las hora a 00:00:00:000
function PrimerDiaMes(AFecha:TDateTime):TDateTime; overload;
function PrimerDiaMes(AMes, Anyo:integer):TDateTime; overload;
//: Ultimo dia del mes y horas a 23:59:59:999
function UltimoDiaMes(AFecha:TDateTime):TDateTime; overload;
function UltimoDiaMes(AMes, Anyo:integer):TDateTime; overload;
//: devuelve el da 15 del mes, para la fecha dada
function Dia15Mes(AFecha:TDateTime):TDateTime;
// Devuelve si en un intevalo de fechas hay das de aos bisiestos
function HayAnyoBisiestoEnIntervalo(DateIni, DateFin:TDate; var ANumDias:integer):boolean;



// Versin (Externa) ===========================================================
function GetFileVersion_(AFileName: string): string;
function VersionToString (version : TULargeInteger) : string;
function StringToVersion (version : string) : TULargeInteger;
// version del programa en formato String
function GetAppVersion(AExeName:string=''; FormatoCorto:Boolean=False): string;


// Codificacion y decodificacion sencilla ======================================
function Decode64(S: string): string;
function Encode64(S: string): string;
function EncodeStrQC(S:string):string;
function DecodeStrQC(S:String):string;

// STRINGS y FICHEROS ==========================================================
{Grabar una cadena a un fichero}
function StringToFile(AStr:WideString; const AFileName:string):boolean;
{Recuperar el contenido de un fichero en una cadena}
function FileToString(const AFileName:string):WideString;
// Rellenar con un caracter
function Replicate(c : char; nLen : integer) : string;
// ajusta a derecha(D) o izquierda(I) (sOrientacion) con el carcter que requiera (sRelleno) suele ser 0 o ''
function Ajusta(sCampo, sOrientacion : string; iLongitud : integer; sRelleno : char) : string;

// Parte derecha en X caracteres
function Right(const S:String; xLen:TStrLen): String;
// trim por la parte derecha
function RTrim(const S:String): String;

// Registro de Windows =========================================================
// Activar clave del registro para ejecutar al inicio
procedure UpdateRegKey(Activate: boolean);


// Ficheros y path de app.  ====================================================
//: devuelve el path para la aplicacin
function GetAppPath():string;
//: Devuelve el nombre de la aplicacin pero con .INI (para configuracion)
function GetAppIniFile():string;

// Funciones para Variants =====================================================
// Obtener el variant como entero (o 0 si es nuLL)
function GetAsInteger(AValue:Variant):Integer;
// Obtener el variant como entero 64(o 0 si es nuLL)
function GetAsInt64(AValue:Variant):Int64;
// Obtener el variant como float (o 0 si es nuLL)
function GetAsFloat(AValue:Variant):double;
// Obtener un variant como currency
function GetAsCurrency(AValue:Variant):currency;
// Obtener un variant como String
function GetAsString(AValue:Variant):string;



//=========================================================================
//
// I M P L E M E N T A T I O N
//
//=========================================================================
implementation

uses
  IniFiles, SysUtils, DateUtils,
{$IFDEF DELPHIXE_UP}
  Forms,
{$ELSE}
  Forms,
{$ENDIF}
  Variants, Registry,
  UCommonConstantes;


// Guarda un array de valores en el fichero de configuracion
procedure SaveOptionValues(AIniFilePath, ASectionName:string;
                           AName, AValue:array of String); overload;
var
  i: Integer;
begin
  with TIniFile.Create(AIniFilePath) do
  try
    for i := 0 to (Length(AName) - 1) do begin
      WriteString(ASectionName, AName[i], AValue[i]);
    end;
  finally
    Free;
  end;
end;

procedure SaveOptionValues(AIniFilePath, ASectionName:string;
                           AName:string; AValue:TStrings); overload;
var
  i: Integer;
  iniF:TIniFile;
begin
  iniF := TIniFile.Create(AIniFilePath);
  try
    for i := 0 to (AValue.Count - 1) do begin
      iniF.WriteString(ASectionName, AName + IntToStr(i), AValue[i]);
    end;
  finally
    iniF.Free;
  end;
end;

// Recupera un valor almacenado...
function RestoreOptionValue(AIniFilePath, ASectionName:string;
                            AName, ADefault:string):String; overload;
begin
  with TiniFile.Create(AIniFilePath) do
  try
    Result := ReadString(ASectionName, AName, Adefault);
  finally
    Free;
  end;
end;

procedure RestoreOptionValue(AIniFilePath, ASectionName:string;
                            AName:string; var outStrings:TStrings);
var
  i:Integer;
  iniF:TIniFile;
  b:boolean;
  str:string;
begin
  // ini
  i := 0;
  iniF := TiniFile.Create(AIniFilePath);
  try
    // Buscar el siguiente valor
    b := iniF.ValueExists(ASectionName, AName + IntToStr(i));
    outStrings.Clear;
    // Mientras hay valores...
    while (b) do begin
      Str := iniF.ReadString(ASectionName, AName + IntToStr(i), STR_EMPTY);
      outStrings.Add(Str);
      // Siguiente
      Inc(i);
      b := iniF.ValueExists(ASectionName, AName + IntToStr(i));
    end;
  finally
    iniF.Free;
  end;
end;

//: Devuelve si est guardada determinada opcin
function OptionExist(AIniFilePath, ASectionName:string; AName:string):Boolean;
begin
  with TiniFile.Create(AIniFilePath) do
  try
    Result := ValueExists(ASectionName, AName)
  finally
    Free;
  end;
end;

// Recupera un valor numerico almacenado...
function RestoreOptionValue(AIniFilePath, ASectionName:string;
                            AName:string; ADefault:Integer): Integer; overload;
begin
  // Llamada con string
  Result := StrToIntDef(RestoreOptionValue(AIniFilePath, ASectionName, AName, IntToStr(ADefault)), ADefault);
end;

function RestoreOptionValue(AIniFilePath, ASectionName:string;
                            AName:string; ADefault:boolean): boolean; overload;
var
  str:string;
begin
  // Llamada con string
  Str := RestoreOptionValue(AIniFilePath, ASectionName, AName, '');
  // convertir
  Result := StrToBoolDef(Str, ADefault);
end;


// devuelve el primer dia del mes y las hora a 0
function PrimerDiaMes(AFecha:TDateTime):TDateTime; overload;
var
  AYear, AMonth, ADay, AHour, AMinute, ASecond, AMilliSecond: Word;
begin
  DecodeDateTime(AFecha, AYear, AMonth, ADay, AHour, AMinute, ASecond, AMilliSecond);
  ADay := 1; AHour := 0; AMinute := 0; ASecond := 0; AMilliSecond := 0;
  Result := EncodeDateTime(AYear, AMonth, ADay, AHour, AMinute, ASecond, AMilliSecond);
end;
// Primer da del mes con hora00:00:00
function PrimerDiaMes(AMes, Anyo:integer):TDateTime; overload;
begin
  Result := EncodeDateTime(Anyo, AMes, 1, 0, 0, 0, 0);
end;

//: Ultimo dia del mes y horas a 23:59:59:999
function UltimoDiaMes(AFecha:TDateTime):TDateTime; overload;
var
  AYear, AMonth, ADay, AHour, AMinute, ASecond, AMilliSecond: Word;
begin
  DecodeDateTime(AFecha, AYear, AMonth, ADay, AHour, AMinute, ASecond, AMilliSecond);
  ADay := 1; AHour := 23; AMinute := 59; ASecond := 59; AMilliSecond := 999;
  // Ultimo dia del mes
  ADay := DaysInAMonth(AYear, AMonth);
  Result := EncodeDateTime(AYear, AMonth, ADay, AHour, AMinute, ASecond, AMilliSecond);
end;

function UltimoDiaMes(AMes, Anyo:integer):TDateTime; overload;
var
  lastDay:Integer;
begin
  lastDay := DaysInAMonth(Anyo, AMes);
  Result := EncodeDateTime(Anyo, AMes, lastDay, 23, 59, 59, 999);
end;

// devuelve el da 15 del mes, para la fecha dada
function Dia15Mes(AFecha:TDateTime):TDateTime;
var
  AYear, AMonth, ADay, AHour, AMinute, ASecond, AMilliSecond: Word;
begin
  DecodeDateTime(AFecha, AYear, AMonth, ADay, AHour, AMinute, ASecond, AMilliSecond);
  ADay := 15; AHour := 0; AMinute := 0; ASecond := 0; AMilliSecond := 0;
  Result := EncodeDateTime(AYear, AMonth, ADay, AHour, AMinute, ASecond, AMilliSecond);
end;

// Devuelve si en un intevalo de fechas hay das de aos bisiestos
function HayAnyoBisiestoEnIntervalo(DateIni, DateFin:TDate; var ANumDias:integer):boolean;
var
  year1, year2:integer;
  i:integer;
  d1, d2:TDate;
begin
  // ini
  ANumDias := 0;
  Result := False;
  year1 := YearOf(DateIni);
  year2 := YearOf(DateFin);
  // recorrido para calcular cuantos "dias 29" tenemos en el intervalo
  for i := year1 to year2 do begin
		if IsLeapYear(i) then begin
			// calcula el bisiesto de cada ao
			d1 := EncodeDate(i, 2, 29);
      // mira si est en el intervalo
      if (d1 >= DateIni) and (d1 <= Datefin) then begin
        Inc(ANumDias);
        Result := True;
      end;
    end;
  end;
end;


function GetFileVersion_(AFileName: string): string;
var
  Size,
    Size2: DWord;
  Pt,
    Pt2: Pointer;
begin
  Size := GetFileVersionInfoSize(PChar(AFileName), Size2);
  if Size > 0 then
  begin
    GetMem(Pt, Size);
    try
      GetFileVersionInfo(PChar(AFileName), 0, Size, Pt);
      VerQueryValue(Pt, '', Pt2, Size2);
      with TVSFixedFileInfo(Pt2^) do
        Result := Format('%d.%d.%d.%d', [HiWord(dwFileVersionMS),
          LoWord(dwFileVersionMS),
            HiWord(dwFileVersionLS),
            LoWord(dwFileVersionLS)]);
    finally
      FreeMem(Pt);
    end;
  end;
end;



// version del programa en formato String
function GetAppVersion(AExeName:string=''; FormatoCorto:Boolean=False): string;
var
  VerInfoSize: DWORD;
  VerInfo: Pointer;
  VerValueSize: DWORD;
  VerValue: PVSFixedFileInfo;
  Dummy: DWORD;
begin
  Result := '';
  if (AExeName = '') then begin
    AExeName := Application.ExeName;
  end;
  VerInfoSize := GetFileVersionInfoSize(PChar(AExeName), Dummy);
  if VerInfoSize = 0 then Exit;
  GetMem(VerInfo, VerInfoSize);
  GetFileVersionInfo(PChar(AExeName), 0, VerInfoSize, VerInfo);
  VerQueryValue(VerInfo, '\', Pointer(VerValue), VerValueSize);
  with VerValue^ do
  begin
    Result := IntToStr(dwFileVersionMS shr 16);
    Result := Result + '.' + IntToStr(dwFileVersionMS and $FFFF);
    // Si no es formato corto, aadimos los 2 ultimos
    if (not FormatoCorto) then begin
      Result := Result + '.' + IntToStr(dwFileVersionLS shr 16);
      Result := Result + '.' + IntToStr(dwFileVersionLS and $FFFF);
    end;
  end;
  FreeMem(VerInfo, VerInfoSize);
end;





(*----------------------------------------------------------------------------*
 | function VersionToString ()                                                |
 |                                                                            |
 | Convert a version large integer to a string                                |
 |                                                                            |
 | Parameters:                                                                |
 |   version : TULargeInteger     The version integer to convert              |
 |                                                                            |
 | The function returns string representation of the version no               |
 *----------------------------------------------------------------------------*)
function VersionToString (version : TULargeInteger) : string;
begin
  with version do
    result := Format ('%d.%d.%d.%d', [HiWord (HighPart), LoWord (HighPart), HiWord (LowPart), LoWord (LowPart)]);
end;

(*----------------------------------------------------------------------------*
 | function StringToVersion ()                                                |
 |                                                                            |
 | Convert a version string to a large integer                                |
 |                                                                            |
 | Parameters:                                                                |
 |   version : string       The version string to convert                     |
 |                                                                            |
 | The function returns the integer representation of the version string      |
 *----------------------------------------------------------------------------*)
function StringToVersion (version : string) : TULargeInteger;
var
  p : Integer;
  s : string;
  hh, h, l, ll : word;
  ok : boolean;
begin
  hh := 0;
  ll := 0;
  h := 0;
  l := 0;

  if (version = STR_EMPTY) then version := ZERO_VERSION;

  s := version;
  p := Pos ('.', s);
  ok := False;
  if p > 0 then
  begin
    hh := StrToInt (Copy (s, 1, p - 1));
    s := Copy (s, p + 1, MaxInt);
    p := Pos ('.', s);
    if p > 0 then
    begin
      h := StrToInt (Copy (s, 1, p - 1));
      s := Copy (s, p + 1, MaxInt);
      p := Pos ('.', s);
      if p > 0 then
      begin
        l := StrToInt (Copy (s, 1, p - 1));
        ll := StrToInt (Copy (s, p + 1, MaxInt));
        ok := True;
      end
    end
  end;

  if not ok then
    raise exception.Create ('Error in version string');

  result.HighPart := 65536 * hh + h;
  result.LowPart := 65536 * l + ll;
end;


// Codificacion
function Encode64(S: string): string;
var
  i: Integer;
  a: Integer;
  x: Integer;
  b: Integer;
begin
  Result := '';
  a := 0;
  b := 0;
  for i := 1 to Length(s) do
  begin
    x := Ord(s[i]);
    b := b * 256 + x;
    a := a + 8;
    while a >= 6 do
    begin
      a := a - 6;
      x := b div (1 shl a);
      b := b mod (1 shl a);
      Result := Result + Codes64[x + 1];
    end;
  end;
  if a > 0 then
  begin
    x := b shl (6 - a);
    Result := Result + Codes64[x + 1];
  end;
end;

// Decodificacion
function Decode64(S: string): string;
var
  i: Integer;
  a: Integer;
  x: Integer;
  b: Integer;
begin
  Result := '';
  a := 0;
  b := 0;
  for i := 1 to Length(s) do
  begin
    x := Pos(s[i], codes64) - 1;
    if x >= 0 then
    begin
      b := b * 64 + x;
      a := a + 6;
      if a >= 8 then
      begin
        a := a - 8;
        x := b shr a;
        b := b mod (1 shl a);
        x := x mod 256;
        Result := Result + chr(x);
      end;
    end
    else
      Exit;
  end;
end;

// Codificar ubna cadena en QC
function EncodeStrQC(S:string):string;
var
  i:integer;
begin
  Result := S;
  for i := 0 to 3 do begin
    Result := Encode64(Result);
  end;
end;
// decode String
function DecodeStrQC(S:String):string;
var
  i:integer;
begin
  Result := S;
  for i := 0 to 3 do begin
    Result := Decode64(Result);
  end;
end;


{Grabar una cadena a un fichero}
function StringToFile(AStr:WideString; const AFileName:string):boolean;
var
  TS:TStringList;
begin
  TS := TStringList.Create();
  try
    try
      TS.Add(AStr);
      TS.SaveToFile(AFileName);
      Result := True;
    except
      Result:= False;
    end;
  finally
    TS.Free;
  end;
end;

{Recuperar el contenido de un fichero en una cadena}
function FileToString(const AFileName:string):WideString;
var
  TS:TStringList;
begin
  // ini
  Result := STR_EMPTY;
  // No existe el fichero?
  if not FileExists(AFileName) then Exit;

  TS := TStringList.Create();
  try
    try
      TS.LoadFromFile(AFileName);
      Result := TS.Text;
    except
      Result:= STR_EMPTY;
    end;
  finally
    TS.Free;
  end;
end;

// replicar un caracter
function Replicate(c: char; nLen:integer) : string;
begin
  Result := stringOfChar(c, nLen);
end;

// ajusta a derecha o izquierda (sOrientacion) con el carcter que requiera (sRelleno) suele ser 0 o ''
function Ajusta(sCampo, sOrientacion : string; iLongitud : integer; sRelleno : char) : string;
begin
  sCampo := Trim(sCampo);
  if uppercase(sOrientacion) = 'I' then
    Result := Replicate(sRelleno,iLongitud - Length(sCampo))+sCampo
  else
    Result := sCampo+Replicate(sRelleno,iLongitud - Length(sCampo));

  if length(Result) > iLongitud then
    Result := copy(Result, 1, iLongitud);
end;

// devuelve el path para la aplicacin
function GetAppPath():string;
begin
  Result := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName));
end;


//: Devuelve el nombre de la aplicacin pero con .INI (para configuracion)
function GetAppIniFile():string;
var
  fName:TFileName;
begin
  fName := ExtractFileName(Application.ExeName);
  fName:= ChangeFileExt(fName, INI_EXT);
  Result := GetAppPath() + fName;
end;

function Right(const S:String; xLen: TStrLen): String;
begin
   Result := Copy( S, (Length( S )-xLen+1 ), xLen );
end;

function RTrim(const S:String): String;
begin
   Result := Trim(S);
end;


// Obtener el variant como entero (o 0 si es nuLL)
function GetAsInteger(AValue:Variant):Integer;
begin
  Result := 0;
  if VarIsNull(AValue) then Exit;
  Result := Integer(AValue);
end;

// Obtener el variant como entero 64(o 0 si es nuLL)
function GetAsInt64(AValue:Variant):Int64;
begin
  Result := 0;
  if VarIsNull(AValue) then Exit;
  Result := Integer(AValue);
end;

// Obtener el variant como float (o 0 si es nuLL)
function GetAsFloat(AValue:Variant):double;
begin
  Result := 0;
  if VarIsNull(AValue) then Exit;
  Result := Double(AValue);
end;

// Obtener el variant como currency (o 0 si es nuLL)
function GetAsCurrency(AValue:Variant):Currency;
begin
  Result := 0;
  if VarIsNull(AValue) then Exit;
  Result := Currency(AValue);
end;

// Obtener un variant como String
function GetAsString(AValue:Variant):string;
begin
  Result := STR_EMPTY;
  if VarIsNull(AValue) then Exit;
  Result := String(AValue);
end;

// ACtualizar la clave del Registro de la propia Aplicacin
procedure UpdateRegKey(Activate: boolean);
const
  KEY_RUN_INI = 'Software\Microsoft\Windows\CurrentVersion\Run';
var
  r:TRegistry;
  b:boolean;
  fName:TFileName;
  KEY_SOFTWARE_NAMEREG:string;      // Nombre del programa
begin
  // ini
  KEY_SOFTWARE_NAMEREG := ExtractFileName(Application.ExeName);
  KEY_SOFTWARE_NAMEREG := ChangeFileExt(KEY_SOFTWARE_NAMEREG, STR_EMPTY);

  // Activar la ejecucin al iniciar el sistema
  r := TRegistry.Create();
  try
    // nombre del jecutable
    fName := Application.ExeName;
    // Existe?
    if FileExists(fName) then begin
      r.RootKey := HKEY_CURRENT_USER;
      b := r.OpenKey(KEY_RUN_INI, False);

      if (not b) then Exit;

      // Marcado?
      if (Activate) then begin
        // Activar
        r.WriteString(KEY_SOFTWARE_NAMEREG, fName);
      end
      else begin
        // Borrar
        b := r.DeleteKey(KEY_SOFTWARE_NAMEREG);
        b := r.DeleteValue(KEY_SOFTWARE_NAMEREG);
      end;
    end;
  finally
    r.Free;
  end;
end;

end.
